/*
 *  OpenVPN-GUI -- A Windows GUI for OpenVPN.
 *
 *  Copyright (C) 2004 Mathias Sundman <mathias@nilings.se>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program (see the file COPYING included with this
 *  distribution); if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Parts of this sourcefile is taken from openvpnserv.c from the
 *  OpenVPN source, with approval from the author, James Yonan
 *  <jim@yonan.net>.
 */


#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <process.h>
#include "tray.h"
#include "main.h"
#include "openvpn.h"
#include "openvpn-gui-res.h"
#include "options.h"
#include "viewlog.h"
#include "proxy.h"
#include "ieproxy.h"
#include <shlobj.h>
#include <richedit.h>

/* These are new for 2000/XP, so they aren't in the mingw headers yet */
#ifndef BELOW_NORMAL_PRIORITY_CLASS
#define BELOW_NORMAL_PRIORITY_CLASS 0x00004000
#endif
#ifndef ABOVE_NORMAL_PRIORITY_CLASS
#define ABOVE_NORMAL_PRIORITY_CLASS 0x00008000
#endif

char passphrase[256];


static bool
match (const WIN32_FIND_DATA *find, const char *ext)
{
  int i;

  if (find->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    return false;

  if (!strlen (ext))
    return true;

  i = strlen (find->cFileName) - strlen (ext) - 1;
  if (i < 1)
    return false;

  return find->cFileName[i] == '.' && !strcmp (find->cFileName + i + 1, ext);
}

/*
 * Modify the extension on a filename.
 */
static bool
modext (char *dest, int size, const char *src, const char *newext)
{
  int i;

  if (size > 0 && (strlen (src) + 1) <= size)
    {
      strcpy (dest, src);
      dest [size - 1] = '\0';
      i = strlen (dest);
      while (--i >= 0)
	{
	  if (dest[i] == '\\')
	    break;
	  if (dest[i] == '.')
	    {
	      dest[i] = '\0';
	      break;
	    }
	}
      if (strlen (dest) + strlen(newext) + 2 <= size)
	{
	  strcat (dest, ".");
	  strcat (dest, newext);
	  return true;
	}
      dest [0] = '\0';
    }
  return false;
}

int
BuildFileList()
{
  WIN32_FIND_DATA find_obj;
  HANDLE find_handle;
  BOOL more_files;
  char find_string[MAX_PATH];
  int i;
  char log_file[MAX_PATH];
  extern struct options o;
  
  mysnprintf (find_string, "%s\\*", o.config_dir);

  find_handle = FindFirstFile (find_string, &find_obj);
  if (find_handle == INVALID_HANDLE_VALUE)
    {
      ShowMsg ("Cannot get configuration file list using: %s", find_string);
      goto finish;
    }

  /*
   * Loop over each config file
   */
  do {
    if (o.num_configs >= MAX_CONFIGS) {
      ShowMsg(GUI_NAME, "OpenVPN GUI does not support more than %d configs. Please contact " \
           "the author if you have the need for more.", MAX_CONFIGS)
      break;
    }

    /* does file have the correct type and extension? */
    if (match (&find_obj, o.ext_string))
      {
        strncpy(o.cnn[o.num_configs].config_file, find_obj.cFileName, sizeof(o.cnn[o.num_configs].config_file));

        /* get log file pathname */
        if (!modext (log_file, sizeof (log_file), o.cnn[o.num_configs].config_file, "log"))
          {
            ShowMsg ("Cannot construct logfile name based on: %s", o.cnn[o.num_configs].config_file);
            goto finish;
          }
        mysnprintf (o.cnn[o.num_configs].log_path, "%s\\%s", o.log_dir, log_file);

        /* Set askpass to true as default */
        o.cnn[o.num_configs].ask_pass = true;

        /* Check if connection should be autostarted */
        for (i=0; (i < MAX_CONFIGS) && o.auto_connect[i]; i++)
          {
            if (strcasecmp(o.cnn[o.num_configs].config_file, o.auto_connect[i]) == 0)
              {
                o.cnn[o.num_configs].auto_connect = true;
                o.cnn[o.num_configs].ask_pass= (o.auto_connect_nopass[i] ? false:true);
                break;
              }
          }

        o.num_configs++;
      }
      
    /* more files to process? */
    more_files = FindNextFile (find_handle, &find_obj);
  } while (more_files);
    
  FindClose (find_handle);
  
  if (o.num_configs == 0) {
    char msg[256];
    ShowMsg(GUI_NAME, "No configuration files were found in %s. " \
                      "Please start OpenVPN-GUI again when you have created one.", \
                      o.config_dir);
    return(false);
  }


  return(true);

finish:
  return(false);
}


int
GetRegistryKeys()
{
  char windows_dir[MAX_PATH];
  char temp_path[MAX_PATH];
  char openvpn_path[MAX_PATH];
  TCHAR programs_dir[MAX_PATH];
  extern struct options o;
  HKEY regkey;

  if (!GetWindowsDirectory(windows_dir, sizeof(windows_dir))) {
    ShowMsg(GUI_NAME, "Error Getting Windows Directory.");
    return(false);
  }
  if(!SUCCEEDED(SHGetFolderPath(NULL, CSIDL_PROGRAM_FILES, NULL, 0, programs_dir))) {
     ShowMsg(GUI_NAME, "Error Getting \"Program\" folder name.");
     return(false);
  }

  /* Get path to OpenVPN installation. */
  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\OpenVPN", 0, KEY_READ, &regkey)
      != ERROR_SUCCESS) 
    {
      ShowMsg(GUI_NAME, "Error opening registry for reading (HKLM\\SOFTWARE\\OpenVPN).\n" \
                        "OpenVPN is probably not installed.");
      return(false);
    }
  if (!GetRegistryValue(regkey, "", openvpn_path, sizeof(openvpn_path)))
    {
      ShowMsg(GUI_NAME, "Error reading registry value (HKLM\\SOFTWARE\\OpenVPN).");
      return(false);
    }
  if(openvpn_path[strlen(openvpn_path) - 1] != '\\')
    strcat(openvpn_path, "\\");


  mysnprintf(temp_path, "%sconfig", openvpn_path);
  if (!GetRegKey("config_dir", o.config_dir, 
      temp_path, sizeof(o.config_dir))) return(false);

  if (!GetRegKey("config_ext", o.ext_string, "ovpn", sizeof(o.ext_string))) return(false);

  mysnprintf(temp_path, "%sbin\\openvpn.exe", openvpn_path);
  if (!GetRegKey("exe_path", o.exe_path, 
      temp_path, sizeof(o.exe_path))) return(false);

  mysnprintf(temp_path, "%slog", openvpn_path);
  if (!GetRegKey("log_dir", o.log_dir, 
      temp_path, sizeof(o.log_dir))) return(false);

  if (!GetRegKey("log_append", o.append_string, "0", sizeof(o.append_string))) return(false);

  if (!GetRegKey("priority", o.priority_string, 
      "NORMAL_PRIORITY_CLASS", sizeof(o.priority_string))) return(false);

  mysnprintf(temp_path, "%s\\notepad.exe", windows_dir);
  if (!GetRegKey("log_viewer", o.log_viewer, 
      temp_path, sizeof(o.log_viewer))) return(false);

  mysnprintf(temp_path, "%s\\notepad.exe", windows_dir);
  if (!GetRegKey("editor", o.editor, 
      temp_path, sizeof(o.editor))) return(false);

  if (!GetRegKey("allow_edit", o.allow_edit, "1", sizeof(o.allow_edit))) return(false);
  
  if (!GetRegKey("allow_service", o.allow_service, "0", sizeof(o.allow_service))) return(false);

  if (!GetRegKey("passphrase_attempts", o.psw_attempts_string, "3", 
      sizeof(o.psw_attempts_string))) return(false);
  o.psw_attempts = atoi(o.psw_attempts_string);
  if ((o.psw_attempts < 1) || (o.psw_attempts > 9))
    {
      ShowMsg(GUI_NAME, "Registry value \"passphrase_attempts\" must be a number between 1 and 9.");
      return(false);
    }

  return(true);
}


int GetRegKey(char name[], char *data, char default_data[], DWORD len)
{
  LONG status;
  DWORD type;
  HKEY openvpn_key;
  HKEY openvpn_key_write;
  DWORD dwDispos;

  /* If option is already set via cmd-line, return */
  if (data[0] != 0) 
    {
      return(true);
    }

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                       "SOFTWARE\\OpenVPN-GUI",
                       0,
                       KEY_READ,
                       &openvpn_key);

  if (status != ERROR_SUCCESS)
    {
      if (RegCreateKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\OpenVPN-GUI",
                        0,
                        "",
                        REG_OPTION_NON_VOLATILE,
                        KEY_READ | KEY_WRITE,
                        NULL,
                        &openvpn_key,
                        &dwDispos) != ERROR_SUCCESS)
        {
          ShowMsg(GUI_NAME, "Error creating HKLM\\SOFTWARE\\OpenVPN-GUI key.");
          return(false);
        }  
    }


  /* get a registry string */
  status = RegQueryValueEx(openvpn_key, name, NULL, &type, (byte *) data, &len);
  if (status != ERROR_SUCCESS || type != REG_SZ)
    {
      /* key did not exist - set default value */
      status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
			  "SOFTWARE\\OpenVPN-GUI",
			  0,
			  KEY_READ | KEY_WRITE,
			  &openvpn_key_write);

      if (status != ERROR_SUCCESS) {
         ShowMsg(GUI_NAME, "Failed to open the registry for writing. " \
             "You need to run this application once as Administrator to update the registry.");
         return(false);
      }    
      if(RegSetValueEx(openvpn_key_write,
                       name,
                       0,
                       REG_SZ,
                       default_data,
                       strlen(default_data)+1))
        {
          ShowMsg(GUI_NAME, "Error reading and setting registry key \"%s\".", name);
          return(false);
        }
      strncpy(data, default_data, len);
      RegCloseKey(openvpn_key_write);

    }
  RegCloseKey(openvpn_key);
  return(true);
}

int StartOpenVPN(int config)
{

  HANDLE hOutputReadTmp = NULL;
  HANDLE hOutputRead = NULL;
  HANDLE hOutputWrite = NULL;
  HANDLE hInputWriteTmp = NULL;
  HANDLE hInputRead = NULL;
  HANDLE hInputWrite = NULL;
  HANDLE hErrorWrite = NULL;

  HANDLE hThread; 
  DWORD IDThread; 
  DWORD priority;
  char exit_event_name[50];
  STARTUPINFO start_info;
  PROCESS_INFORMATION proc_info;
  SECURITY_ATTRIBUTES sa;
  SECURITY_DESCRIPTOR sd;
  char command_line[256];
  char proxy_string[100];
  char msg[200];
  int i, is_connected=0;

  extern HINSTANCE hInstance;
  extern struct options o;


  CLEAR (start_info);
  CLEAR (proc_info);
  CLEAR (sa);
  CLEAR (sd);

  /* Clear failed because of wrong psw flag */
  o.cnn[config].failed_psw = 0;


  /* If oldversion, allow only ONE connection */
  if (o.oldversion == 1)
    {
      for (i=0; i < o.num_configs; i++) {
        if (o.cnn[i].connect_status != 0) {
          is_connected=1;
          break;
        }
      }
     if (is_connected) {
       ShowMsg(GUI_NAME, "You can only have one connection running at " \
                         "the same time when using an older version " \
                         "of OpenVPN than 2.0-beta6.");
        return(false);
      }
    }

  /*
   * Create our exit event
   */
  o.cnn[config].exit_event = NULL;
  if (o.oldversion == 1)
    {
      mysnprintf(exit_event_name, "openvpn_exit");
    }
  else
    {
      mysnprintf(exit_event_name, "openvpn_exit_event_%d",config);
    }
  o.cnn[config].exit_event = CreateEvent (NULL, TRUE, FALSE, exit_event_name);
  if (o.cnn[config].exit_event == NULL)
    {
      ShowMsg (GUI_NAME, "CreateEvent failed on exit event: %s", exit_event_name);
      return(false);
    }

  /* set process priority */
  priority = NORMAL_PRIORITY_CLASS;
  if (!strcmp (o.priority_string, "IDLE_PRIORITY_CLASS"))
    priority = IDLE_PRIORITY_CLASS;
  else if (!strcmp (o.priority_string, "BELOW_NORMAL_PRIORITY_CLASS"))
    priority = BELOW_NORMAL_PRIORITY_CLASS;
  else if (!strcmp (o.priority_string, "NORMAL_PRIORITY_CLASS"))
    priority = NORMAL_PRIORITY_CLASS;
  else if (!strcmp (o.priority_string, "ABOVE_NORMAL_PRIORITY_CLASS"))
    priority = ABOVE_NORMAL_PRIORITY_CLASS;
  else if (!strcmp (o.priority_string, "HIGH_PRIORITY_CLASS"))
    priority = HIGH_PRIORITY_CLASS;
  else
    {
      ShowMsg (GUI_NAME, "Unknown priority name: %s", o.priority_string);
      goto failed;
    }

  /* Check that log append flag has a valid value */
  if ((o.append_string[0] != '0') && (o.append_string[0] != '1'))
    {
      ShowMsg (GUI_NAME, "Log file append flag (given as '%s') must be '0' or '1'", o.append_string);
      goto failed;
    }

      
  /* construct proxy string to append to command line */
  CLEAR(proxy_string);
  if (o.proxy_source == PROXY_SOURCE_MANUAL)
    {
      if (o.proxy_type == PROXY_TYPE_HTTP)
        {
          if (o.proxy_http_auth == PROXY_HTTP_ASK_AUTH)
            {
              /* Ask for Proxy username/password */
              DialogBox(o.hInstance, (LPCTSTR)IDD_PROXY_AUTH, NULL, (DLGPROC)ProxyAuthDialogFunc);
              mysnprintf(proxy_string, "--http-proxy %s %s %s",
                         o.proxy_http_address,
                         o.proxy_http_port,
                         o.proxy_authfile);
            }
          else
            {
              mysnprintf(proxy_string, "--http-proxy %s %s",
                         o.proxy_http_address,
                         o.proxy_http_port);
            }
        }
      if (o.proxy_type == PROXY_TYPE_SOCKS)
        {
          mysnprintf(proxy_string, "--socks-proxy %s %s",
                     o.proxy_socks_address,
                     o.proxy_socks_port);
        }
    }

  if (o.proxy_source == PROXY_SOURCE_IE)
    {
      LPTSTR ieproxy;
      extern LPTSTR getIeHttpProxyError;
      char *pos;

      ieproxy = getIeHttpProxy();
      if (!ieproxy)
        {
          ShowMsg(GUI_NAME, "Unable to get IE proxy settings because:\n\n%s",
                  getIeHttpProxyError);
        }
      else
        {
          /* Don't use a proxy if IE proxy string is empty */
          if (strlen(ieproxy) > 1)
            {
              /* Parse the port number from the IE proxy string */
              if ((pos = strchr(ieproxy, ':')))
                {
                  *pos = '\0';
                  if (o.proxy_http_auth == PROXY_HTTP_ASK_AUTH)
                    {
                      /* Ask for Proxy username/password */
                      DialogBox(o.hInstance, (LPCTSTR)IDD_PROXY_AUTH, NULL, (DLGPROC)ProxyAuthDialogFunc);
                      mysnprintf(proxy_string, "--http-proxy %s %s %s", 
                                 ieproxy, pos+1, o.proxy_authfile)
                    }
                  else
                    {
                      mysnprintf(proxy_string, "--http-proxy %s %s", ieproxy, pos+1)
                    }
                }
              /* No port is specified, use 80 as default port */
              else
                {
                  if (o.proxy_http_auth == PROXY_HTTP_ASK_AUTH)
                    {
                      /* Ask for Proxy username/password */
                      DialogBox(o.hInstance, (LPCTSTR)IDD_PROXY_AUTH, NULL, (DLGPROC)ProxyAuthDialogFunc);
                      mysnprintf(proxy_string, "--http-proxy %s 80 %s", 
                                 ieproxy, o.proxy_authfile)
                    }
                  else
                    {
                      mysnprintf(proxy_string, "--http-proxy %s 80", ieproxy)
                    }
                }
            }
          free (ieproxy);
        }
    }
      
  /* construct command line */
  if (o.oldversion == 1)
    {
      mysnprintf (command_line, "openvpn --config \"%s\" %s",
                  o.cnn[config].config_file, proxy_string);
    }
  else
    {
      mysnprintf (command_line, "openvpn --service %s 1 --config \"%s\" %s",
                  exit_event_name,
                  o.cnn[config].config_file,
                  proxy_string);
    }


  /* Make security attributes struct for logfile handle so it can
     be inherited. */
  sa.nLength = sizeof (sa);
  sa.lpSecurityDescriptor = &sd;
  sa.bInheritHandle = TRUE;
  if (!InitializeSecurityDescriptor (&sd, SECURITY_DESCRIPTOR_REVISION))
    {
      ShowMsg (GUI_NAME, "InitializeSecurityDescriptor failed");
      goto failed;
    }
  if (!SetSecurityDescriptorDacl (&sd, TRUE, NULL, FALSE))
    {
      ShowMsg (GUI_NAME, "SetSecurityDescriptorDacl failed");
      goto failed;
    }


  /* Create the child output pipe. */
  if (!CreatePipe(&hOutputReadTmp,&hOutputWrite,&sa,0))
    {
      ShowMsg(GUI_NAME, "CreatePipe on hOutputWrite failed.");
      goto failed;
    }

  // Create a duplicate of the output write handle for the std error
  // write handle. This is necessary in case the child application
  // closes one of its std output handles.
  if (!DuplicateHandle(GetCurrentProcess(),hOutputWrite,
                       GetCurrentProcess(),&hErrorWrite,0,
                       TRUE,DUPLICATE_SAME_ACCESS))
    { 
      ShowMsg(GUI_NAME, "DuplicateHandle on hErrorWrite failed.");
      goto failed;
    }

  // Create the child input pipe.
  if (!CreatePipe(&hInputRead,&hInputWriteTmp,&sa,0))
    {
      ShowMsg(GUI_NAME, "CreatePipe on hInputRead failed.");
      goto failed;
    }

  // Create new output read handle and the input write handles. Set
  // the Properties to FALSE. Otherwise, the child inherits the
  // properties and, as a result, non-closeable handles to the pipes
  // are created.
  if (!DuplicateHandle(GetCurrentProcess(),hOutputReadTmp,
                       GetCurrentProcess(),
                       &hOutputRead, // Address of new handle.
                       0,FALSE, // Make it uninheritable.
                       DUPLICATE_SAME_ACCESS))
    {
      ShowMsg(GUI_NAME, "DupliateHandle on hOutputRead failed.");
      goto failed;
    }

  if (!DuplicateHandle(GetCurrentProcess(),hInputWriteTmp,
                       GetCurrentProcess(),
                       &hInputWrite, // Address of new handle.
                       0,FALSE, // Make it uninheritable.
                       DUPLICATE_SAME_ACCESS))
    {
      ShowMsg(GUI_NAME, "DupliateHandle on hInputWrite failed.");
      goto failed;
    }

  /* Close inheritable copies of the handles */
  if (!CloseHandle(hOutputReadTmp) || !CloseHandle(hInputWriteTmp)) 
    {
      ShowMsg(GUI_NAME, "CloseHandle on hOutputReadTmp/hInputWriteTmp failed.");
      CloseHandle (o.cnn[config].exit_event);
      return(0); 
    }
  hOutputReadTmp=NULL;
  hInputWriteTmp=NULL;

  /* fill in STARTUPINFO struct */
  GetStartupInfo(&start_info);
  start_info.cb = sizeof(start_info);
  start_info.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
  start_info.wShowWindow = SW_HIDE;
  start_info.hStdInput = hInputRead;
  start_info.hStdOutput = hOutputWrite;
  start_info.hStdError = hErrorWrite;

  /* create an OpenVPN process for one config file */
  if (!CreateProcess(o.exe_path,
		     command_line,
		     NULL,
		     NULL,
		     TRUE,
		     priority | CREATE_NEW_CONSOLE,
		     NULL,
		     o.config_dir,
		     &start_info,
		     &proc_info))
    {
      ShowMsg (GUI_NAME, "CreateProcess failed, exe='%s' cmdline='%s' dir='%s'",
          o.exe_path,
          command_line,
          o.config_dir);
      goto failed;
    }


  /* close unneeded handles */
  Sleep (250); /* try to prevent race if we close logfile
                  handle before child process DUPs it */

  if(!CloseHandle (proc_info.hThread) ||
     !CloseHandle (hOutputWrite) ||
     !CloseHandle (hInputRead) ||
     !CloseHandle (hErrorWrite))
    {
      ShowMsg (GUI_NAME, "CloseHandle failed");
      CloseHandle (o.cnn[config].exit_event);
      return(false);
    }
  hOutputWrite = NULL;
  hInputRead = NULL;
  hErrorWrite = NULL;
 
  /* Save StdIn and StdOut handles in our options struct */
  o.cnn[config].hStdIn = hInputWrite;
  o.cnn[config].hStdOut = hOutputRead;

  /* Save Process Handle */
  o.cnn[config].hProcess=proc_info.hProcess;


  /* Start Thread to show Status Dialog */
  hThread = CreateThread(NULL, 0, 
            (LPTHREAD_START_ROUTINE) ThreadOpenVPNStatus,
            (int *) config,  // pass config nr
            0, &IDThread); 
  if (hThread == NULL) 
    {
    ShowMsg (GUI_NAME, "CreateThread to show Status window Failed.");
    goto failed;
  }


  return(true); 

failed:
  if (o.cnn[config].exit_event) CloseHandle (o.cnn[config].exit_event);
  if (hOutputWrite) CloseHandle (hOutputWrite);
  if (hOutputRead) CloseHandle (hOutputRead);
  if (hInputWrite) CloseHandle (hInputWrite);
  if (hInputRead) CloseHandle (hInputRead);
  if (hErrorWrite) CloseHandle (hOutputWrite);
  return(false);

}


void StopOpenVPN(int config)
{
  int i;
  extern struct options o;

  o.cnn[config].connect_status=false;
  
  if (o.cnn[config].exit_event) {
    EnableWindow(GetDlgItem(o.cnn[config].hwndStatus, ID_DISCONNECT), FALSE);
    SetMenuStatus(config, 3);
    SetDlgItemText(o.cnn[config].hwndStatus, TEXT_STATUS, 
                   "Current State: Waiting for OpenVPN to terminate..."); 
    SetEvent(o.cnn[config].exit_event);
  }
}

void StopAllOpenVPN()
{
  int i;
  extern struct options o;
 
  for(i=0; i < o.num_configs; i++) {
    if(o.cnn[i].connect_status != 0)
      StopOpenVPN(i);
  }

  Sleep(2000);
}


void WatchOpenVPNProcess(int config)
{
  char line[1024];
  extern struct options o;
  int ret;
  char filemode[2] = "w\0";
  FILE *fd;
  char conn_name[200];
  char msg[200];
  int i;
  int iLineCount;
  int LogLines = 0;
  int logpos;

  HWND LogWindow;

  /* set log file append/truncate filemode */
  if (o.append_string[0] == '1')
    filemode[0] = 'a';

  /* Cut of extention from config filename. */
  strncpy(conn_name, o.cnn[config].config_file, sizeof(conn_name));
  conn_name[strlen(conn_name) - (strlen(o.ext_string)+1)]=0;

  /* Set Connect_Status = "Connecting" */
  o.cnn[config].connect_status=1;
 
  /* Set Tray Icon = "Connecting" if no other connections are running */
  if (CountConnectedState(2) == 0)
    SetTrayIcon(1);

  /* Set MenuStatus = "Connecting" */
  SetMenuStatus(config, 1);
  
  /* Clear failed_password flag */
  o.cnn[config].failed_psw = 0;

  /* Open log file */
  if ((fd=fopen(o.cnn[config].log_path, filemode)) == NULL)
    {
      ShowMsg (GUI_NAME, "Error openvpn logfile for writing: %s", o.cnn[config].log_path);
      return;
    }

  LogWindow = GetDlgItem(o.cnn[config].hwndStatus, EDIT_LOG);
  while(TRUE)
    {
      if ((ret=ReadLineFromStdOut(o.cnn[config].hStdOut, config, line)) == 1)
        {

          /* Do nothing if line length = 0 */
          if (strlen(line) == 0) continue;

          /* Write line to Log file */
          fputs (line, fd);
          fputc ('\n', fd);
          fflush (fd);

          /* Remove lines from LogWindow if it is getting full */
          LogLines++;
          if (LogLines > MAX_LOG_LINES)
            {
              logpos = SendMessage(LogWindow, EM_LINEINDEX, DEL_LOG_LINES, 0);
              SendMessage(LogWindow, EM_SETSEL, 0, logpos);
              SendMessage(LogWindow, EM_REPLACESEL, FALSE, (WPARAM) "");
              LogLines -= DEL_LOG_LINES;
            }

          /* Write line to LogWindow */
          strcat(line, "\r\n");
          SendMessage(LogWindow, EM_SETSEL, -1, -1);
          SendMessage(LogWindow, EM_REPLACESEL, FALSE, (WPARAM) line);

          if (o.cnn[config].connect_status == 1) /* Connecting state */
            {
              /* Check for Connected message */
              if (strstr(line, o.connect_string) != NULL)
                {
                  o.cnn[config].connect_status=2;
                  SetMenuStatus(config, 2);
                  SetTrayIcon(2);

                  /* Remove Proxy Auth file */
                  DeleteFile(o.proxy_authfile);

                  /* Zero psw attempt counter */
                  o.cnn[config].failed_psw_attempts = 0;

                  SetDlgItemText(o.cnn[config].hwndStatus, TEXT_STATUS, "Current State: CONNECTED"); 
                  SetStatusWinIcon(o.cnn[config].hwndStatus, APP_ICON_CONNECTED);

                  ShowMsg(GUI_NAME, "Connection established to %s.", conn_name);

                  /* Hide Status Window */
                  ShowWindow(o.cnn[config].hwndStatus, SW_HIDE);
                }

              /* Check for failed passphrase log message */
              if ((strstr(line, "TLS Error: Need PEM pass phrase for private key") != NULL) ||
                   (strstr(line, "EVP_DecryptFinal:bad decrypt") != NULL) ||
                   (strstr(line, "PKCS12_parse:mac verify failure") != NULL))
                {
                  o.cnn[config].failed_psw_attempts++;
                  o.cnn[config].failed_psw=1;
                }

              /* Check for "certificate has expired" message */
              if ((strstr(line, "error=certificate has expired") != NULL))
                {
                  StopOpenVPN(config);
                  ShowMsg(GUI_NAME, "Unable to connect because your certificate " \
                                    "has expired or the system time is incorrect.");
                }

              /* Check for "certificate is not yet valid" message */
              if ((strstr(line, "error=certificate is not yet valid") != NULL))
                {
                  StopOpenVPN(config);
                  ShowMsg(GUI_NAME, "Unable to connect because your certificate " \
                                    "is not yet valid. Check that your system time is correct.");
                }
            }

          if (o.cnn[config].connect_status == 2) /* Connected state */
            {
              /* Check for Ping-Restart message */
              if (strstr(line, "process restarting") != NULL)
                {
                  /* Set connect_status = ReConnecting */
                  o.cnn[config].connect_status=3;
                  CheckAndSetTrayIcon();

                  /* Set Status Window Controls "ReConnecting" */
                  SetDlgItemText(o.cnn[config].hwndStatus, TEXT_STATUS, "Current State: ReConnecting"); 
                  SetStatusWinIcon(o.cnn[config].hwndStatus, APP_ICON_CONNECTING);
                }
            }

          if (o.cnn[config].connect_status == 3) /* ReConnecting state */
            {
              /* Check for Connected message */
              if (strstr(line, o.connect_string) != NULL)
                {
                  o.cnn[config].connect_status=2;
                  SetTrayIcon(2);

                  /* Set Status Window Controls "Connected" */
                  SetDlgItemText(o.cnn[config].hwndStatus, TEXT_STATUS, "Current State: Connected"); 
                  SetStatusWinIcon(o.cnn[config].hwndStatus, APP_ICON_CONNECTED);
                }

              /* Check for failed passphrase log message */
              if ((strstr(line, "TLS Error: Need PEM pass phrase for private key") != NULL) ||
                   (strstr(line, "EVP_DecryptFinal:bad decrypt") != NULL) ||
                   (strstr(line, "PKCS12_parse:mac verify failure") != NULL))
                {
                  o.cnn[config].failed_psw_attempts++;
                  o.cnn[config].failed_psw=1;
                }
            }
        }
      else break;

    }


  /* OpenVPN has been shutdown */
 
  /* Close logfile filedesc. */
  fclose(fd);

  /* Close StdIn/StdOut handles */
  CloseHandle(o.cnn[config].hStdIn);
  CloseHandle(o.cnn[config].hStdOut);

  /* Close exitevent handle */
  CloseHandle(o.cnn[config].exit_event);
  o.cnn[config].exit_event = NULL;    

  /* Enable/Disable menuitems for this connections */
  SetMenuStatus(config, false);

  /* Remove Proxy Auth file */
  DeleteFile(o.proxy_authfile);

  /* Process died outside our control */
  if(o.cnn[config].connect_status == 2)		
    {
      /* Zero psw attempt counter */
      o.cnn[config].failed_psw_attempts = 0;

      /* Set connect_status = "Not Connected" */
      o.cnn[config].connect_status=0;

      /* Change tray icon if no more connections is running */
      CheckAndSetTrayIcon();

      /* Show Status Window */
      SetDlgItemText(o.cnn[config].hwndStatus, TEXT_STATUS, "Current State: DISCONNECTED"); 
      SetStatusWinIcon(o.cnn[config].hwndStatus, APP_ICON_DISCONNECTED);
      EnableWindow(GetDlgItem(o.cnn[config].hwndStatus, IDCANCEL), FALSE);
      SetForegroundWindow(o.cnn[config].hwndStatus);
      ShowWindow(o.cnn[config].hwndStatus, SW_SHOW);
      ShowMsg(GUI_NAME, "Connection to %s was terminated.", conn_name); 

      /* Close Status Window */
      SendMessage(o.cnn[config].hwndStatus, WM_CLOSE, 0, 0);
    }

  /* We have failed to connect */
  else if(o.cnn[config].connect_status == 1)	
    {

      /* Set connect_status = "Not Connected" */
      o.cnn[config].connect_status=false;

      /* Change tray icon if no more connections is running */
      CheckAndSetTrayIcon();

      if ((o.cnn[config].failed_psw) && 
          (o.cnn[config].failed_psw_attempts < o.psw_attempts))
        {

          /* Close Status Window */
          SendMessage(o.cnn[config].hwndStatus, WM_CLOSE, 0, 0);

          /* Restart OpenVPN to make another attempt to connect */
          PostMessage(o.hWnd, WM_COMMAND, IDM_CONNECTMENU + config, 0);
        }
      else
        {
          /* Show Status Window */
          SetDlgItemText(o.cnn[config].hwndStatus, TEXT_STATUS, "Current State: Failed to connect"); 
          SetStatusWinIcon(o.cnn[config].hwndStatus, APP_ICON_DISCONNECTED);
          EnableWindow(GetDlgItem(o.cnn[config].hwndStatus, IDCANCEL), FALSE);
          SetForegroundWindow(o.cnn[config].hwndStatus);
          ShowWindow(o.cnn[config].hwndStatus, SW_SHOW);

          /* Zero psw attempt counter */
          o.cnn[config].failed_psw_attempts = 0;

          ShowMsg(GUI_NAME, "Connecting to %s has failed.", conn_name);

          /* Close Status Window */
          SendMessage(o.cnn[config].hwndStatus, WM_CLOSE, 0, 0);
        }
    }

  /* We have failed to reconnect */
  else if(o.cnn[config].connect_status == 3)	
    {

      /* Set connect_status = "Not Connected" */
      o.cnn[config].connect_status=false;

      /* Change tray icon if no more connections is running */
      CheckAndSetTrayIcon();

      if ((o.cnn[config].failed_psw) && 
          (o.cnn[config].failed_psw_attempts < o.psw_attempts))
        {

          /* Close Status Window */
          SendMessage(o.cnn[config].hwndStatus, WM_CLOSE, 0, 0);

          /* Restart OpenVPN to make another attempt to connect */
          PostMessage(o.hWnd, WM_COMMAND, IDM_CONNECTMENU + config, 0);
        }
      else
        {
          /* Show Status Window */
          SetDlgItemText(o.cnn[config].hwndStatus, TEXT_STATUS, "Current State: Failed to reconnect"); 
          SetStatusWinIcon(o.cnn[config].hwndStatus, APP_ICON_DISCONNECTED);
          EnableWindow(GetDlgItem(o.cnn[config].hwndStatus, IDCANCEL), FALSE);
          SetForegroundWindow(o.cnn[config].hwndStatus);
          ShowWindow(o.cnn[config].hwndStatus, SW_SHOW);

          /* Zero psw attempt counter */
          o.cnn[config].failed_psw_attempts = 0;

          ShowMsg(GUI_NAME, "ReConnecting to %s has failed.", conn_name);

          /* Close Status Window */
          SendMessage(o.cnn[config].hwndStatus, WM_CLOSE, 0, 0);
        }
    }

  /* We have chosed to Disconnect */
  else 						
    {
      /* Zero psw attempt counter */
      o.cnn[config].failed_psw_attempts = 0;

      /* Set connect_status = "Not Connected" */
      o.cnn[config].connect_status=false;

      /* Change tray icon if no more connections is running */
      CheckAndSetTrayIcon();

      /* Close Status Window */
      SendMessage(o.cnn[config].hwndStatus, WM_CLOSE, 0, 0);
    }

}


/* Wait for a complete line (CR/LF) and return it.
 * Return values:
 *  1 - Successful. Line is available in *line.
 *  0 - Broken Pipe during ReadFile.
 * -1 - Other Error during ReadFile.
 * 
 * I'm really unhappy with this code! If anyone knows of an easier
 * way to convert the streaming data from ReadFile() into lines,
 * please let me know!
 */
int ReadLineFromStdOut(HANDLE hStdOut, int config, char *line)
{
  #define MAX_LINELEN 1024

      CHAR lpBuffer[MAX_LINELEN];
      static char lastline[MAX_CONFIGS][MAX_LINELEN];
      static int charsleft[MAX_CONFIGS];
      char tmpline[MAX_LINELEN];
      DWORD nBytesRead;
      DWORD nCharsWritten;
      char *p;
      int len, i;
      static int first_call = 1;
      extern struct options o;
      extern HINSTANCE hInstance;

  if (first_call)
    {
      for (i=0; i < MAX_CONFIGS; i++)
        charsleft[i]=0;
      first_call = 0;
    }

  while (true)
    {
      if (charsleft[config])
        {
          /* Check for Passphrase prompt */
          if (strncmp(lastline[config], "Enter PEM pass phrase:", 22) == 0)
            {
              CLEAR(passphrase);  
              DialogBox(o.hInstance, (LPCTSTR)IDD_PASSPHRASE, NULL, (DLGPROC)PassphraseDialogFunc);

              if (passphrase)
                {
                  if (strlen(passphrase) > 0)
                    {
                      if (!WriteFile(o.cnn[config].hStdIn, passphrase,
                                strlen(passphrase), &nCharsWritten, NULL))
                        {
                          ShowMsg(GUI_NAME, "Error passing passphrase to StdIn.");
                        }
                    }
                }
              if (!WriteFile(o.cnn[config].hStdIn, "\r\n",
                            2, &nCharsWritten, NULL))
                {
                  ShowMsg(GUI_NAME, "Error passing CR after passphrase to StdIn.");
                }

              /* Remove Passphrase prompt from lastline buffer */
              lastline[config][0]='\0';

            }
         

          p=strchr(lastline[config], '\n');
          if (p == NULL)
            {
              if (!ReadFile(hStdOut,lpBuffer,sizeof(lpBuffer) - 1,
                                              &nBytesRead,NULL) || !nBytesRead)
                {
                  if (GetLastError() == ERROR_BROKEN_PIPE)
                    return(0); // pipe done - normal exit path.
                  else
                    {
                      ShowMsg(GUI_NAME, "Error reading from OpenVPN StdOut Pipe.");
                      return(-1);
                    } 
                }
              lpBuffer[nBytesRead] = '\0';
              p=strchr(lpBuffer, '\n');
              if (p == NULL)
                {
                  strncat(lastline[config], lpBuffer, sizeof(lastline[config]) - strlen(lastline[config]) - 1);
                }
              else
                {
                  p[0] = '\0';
                  strncpy(line, lastline[config], MAX_LINELEN - 1);
                  strncat(line, lpBuffer, MAX_LINELEN - strlen(line) - 1);
                  if (line[strlen(line) - 1] == '\r') line[strlen(line) - 1] = '\0';
                  if (nBytesRead > (strlen(lpBuffer) + 1))
                    {
                      strncpy(lastline[config], p+1, sizeof(lastline[config]) - 1);
                      charsleft[config]=1;
                      return(1);
                    }
                  charsleft[config]=0;
                  return(1);
                }
            }
          else
            {
              len = strlen(lastline[config]);
              p[0] = '\0';
              strncpy(line, lastline[config], MAX_LINELEN - 1);
              if (line[strlen(line) - 1] == '\r') line[strlen(line) - 1] = '\0';
              if (len > (strlen(line) + 2))
                {
                  strncpy(tmpline, p+1, sizeof(tmpline) - 1);
                  strncpy(lastline[config], tmpline, sizeof(lastline[config]) - 1);
                  charsleft[config]=1;
                  return(1);
                }
              charsleft[config]=0;
              return(1);
            }
        }
      else
        {  
          if (!ReadFile(hStdOut,lpBuffer,sizeof(lpBuffer) - 1,
                                          &nBytesRead,NULL) || !nBytesRead)
            {
              if (GetLastError() == ERROR_BROKEN_PIPE)
                return(0); // pipe done - normal exit path.
              else
                {
                  ShowMsg(GUI_NAME, "Error reading from OpenVPN StdOut Pipe.");
                  return(-1);
                } 
            }
          lpBuffer[nBytesRead] = '\0';
          p=strchr(lpBuffer, '\n');
          if (p == NULL)
            {
              strncpy(lastline[config], lpBuffer, sizeof(lastline[config]) - 1);
              charsleft[config]=1;
            }
          else
            {
              p[0] = '\0';
              strncpy(line, lpBuffer, MAX_LINELEN - 1);
              if (line[strlen(line) - 1] == '\r') line[strlen(line) - 1] = '\0';
              if (nBytesRead > strlen(line))
                {
                  strncpy(lastline[config], p+1, sizeof(lastline[config]) - 1);
                  charsleft[config]=1;
                  return(1);
                }
            }
        }     
    }

}

BOOL CALLBACK PassphraseDialogFunc (HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch (msg) {

    case WM_INITDIALOG:
      SetForegroundWindow(hwndDlg);
      break;

    case WM_COMMAND:
      switch (LOWORD(wParam)) {

        case IDOK:			// button
          GetDlgItemText(hwndDlg, EDIT_PASSPHRASE, passphrase, sizeof(passphrase) - 1);
          EndDialog(hwndDlg, LOWORD(wParam));
          return TRUE;
      }
      break;
    case WM_CLOSE:
      EndDialog(hwndDlg, LOWORD(wParam));
      return TRUE;
     
  }
  return FALSE;
}

BOOL CALLBACK StatusDialogFunc (HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch (msg) {
  BOOL Translated;
  HWND hwndLogWindow;
  RECT rect;
  CHARFORMAT charformat;
  HICON hIcon;
  int config;

  extern struct options o;

    case WM_INITDIALOG:
      /* Set Window Icon "DisConnected" */
      SetStatusWinIcon(hwndDlg, APP_ICON_CONNECTING);

      /* Create LogWindow */
      hwndLogWindow = CreateWindowEx (0, RICHEDIT_CLASS, NULL, 
                                      WS_CHILD | WS_VISIBLE | WS_HSCROLL | WS_VSCROLL | \
                                      ES_SUNKEN | ES_LEFT | ES_MULTILINE | \
                                      ES_READONLY | ES_AUTOHSCROLL | ES_AUTOVSCROLL,
                                      20, 25, 350, 160,		// Posision and Size
                                      hwndDlg,			// Parent window handle
                                      (HMENU) EDIT_LOG,		// hMenu
                                      o.hInstance,		// hInstance
                                      NULL);			// WM_CREATE lpParam

                                     
      if (!hwndLogWindow)
        {
          ShowMsg(GUI_NAME, "Creating RichEdit LogWindow Failed!!!");
          return FALSE;
        }

      /* Set font and fontsize of the LogWindow */
      charformat.cbSize = sizeof(CHARFORMAT); 
      charformat.dwMask = CFM_SIZE | CFM_FACE | CFM_BOLD | CFM_ITALIC | \
                          CFM_UNDERLINE | CFM_STRIKEOUT | CFM_PROTECTED;
      charformat.dwEffects = 0;
      charformat.yHeight = 100;
      strcpy(charformat.szFaceName, "MS Sans Serif");
      if ((SendMessage(hwndLogWindow, EM_SETCHARFORMAT, SCF_DEFAULT, (LPARAM) &charformat) && CFM_SIZE) == 0)
        ShowMsg(GUI_NAME, "Set Size failed!");

      /* Set Size and Posision of controls */
      GetClientRect(hwndDlg, &rect);
      MoveWindow (hwndLogWindow, 20, 25, rect.right - 40, rect.bottom - 70, TRUE);
      MoveWindow (GetDlgItem(hwndDlg, TEXT_STATUS), 20, 5, rect.right - 25, 15, TRUE);
      MoveWindow (GetDlgItem(hwndDlg, ID_DISCONNECT), rect.right / 2 - 90, rect.bottom - 30, 80, 23, TRUE);
      MoveWindow (GetDlgItem(hwndDlg, ID_HIDE), rect.right / 2 + 10, rect.bottom - 30, 80, 23, TRUE);

      /* Set focus on the LogWindow so it scrolls automatically */
      SetFocus(hwndLogWindow);

      return FALSE;

    case WM_SIZE:
      MoveWindow (GetDlgItem(hwndDlg, EDIT_LOG), 20, 25, LOWORD (lParam) - 40,
                  HIWORD (lParam) - 70, TRUE);
      MoveWindow (GetDlgItem(hwndDlg, ID_DISCONNECT), LOWORD (lParam) / 2 - 90,
                  HIWORD (lParam) - 30, 80, 23, TRUE);
      MoveWindow (GetDlgItem(hwndDlg, ID_HIDE), LOWORD (lParam) / 2 + 10,
                  HIWORD (lParam) - 30, 80, 23, TRUE);
      MoveWindow (GetDlgItem(hwndDlg, TEXT_STATUS), 20, 5, LOWORD (lParam) - 25, 15, TRUE);
      InvalidateRect(hwndDlg, NULL, TRUE);
      return FALSE;

    case WM_COMMAND:
      switch (LOWORD(wParam)) {

        case ID_DISCONNECT:
          config=GetDlgItemInt(hwndDlg, TEXT_CONFIG, &Translated, FALSE);
          StopOpenVPN(config);
          return TRUE;

        case ID_HIDE:
          if (o.cnn[GetDlgItemInt(hwndDlg, TEXT_CONFIG, &Translated, FALSE)].connect_status != 0)
            {
              ShowWindow(hwndDlg, SW_HIDE);
            }
           else
            {
              DestroyWindow(hwndDlg);
            }
          return TRUE;
      }
      break;

    case WM_CLOSE:
      if (o.cnn[GetDlgItemInt(hwndDlg, TEXT_CONFIG, &Translated, FALSE)].connect_status != 0)
        {
          ShowWindow(hwndDlg, SW_HIDE);
        }
      else
        {
          DestroyWindow(hwndDlg);
        }
      return TRUE;

    case WM_DESTROY:
      PostQuitMessage(0);
      break; 
  }
  return FALSE;
}

void SetStatusWinIcon(HWND hwndDlg, int IconID)
{
  HICON hIcon;

  /* Set Window Icon */
  hIcon = (HICON)LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IconID), 
                                                IMAGE_ICON, 0, 0, LR_DEFAULTCOLOR);
  if (hIcon) {
    SendMessage(hwndDlg, WM_SETICON, (WPARAM) (ICON_SMALL), (LPARAM) (hIcon));
    SendMessage(hwndDlg, WM_SETICON, (WPARAM) (ICON_BIG), (LPARAM) (hIcon));  
  }
}

int AutoStartConnections()
{
  extern struct options o;
  int i;

  for (i=0; i < o.num_configs; i++)
    {
      if (o.cnn[i].auto_connect)
        StartOpenVPN(i);
    }

  return(true);
}

int CheckVersion()
{
  HANDLE hOutputReadTmp = NULL;
  HANDLE hOutputRead = NULL;
  HANDLE hOutputWrite = NULL;
  HANDLE hInputWriteTmp = NULL;
  HANDLE hInputRead = NULL;
  HANDLE hInputWrite = NULL;
  HANDLE exit_event;
  HANDLE debug_event;

  STARTUPINFO start_info;
  PROCESS_INFORMATION proc_info;
  SECURITY_ATTRIBUTES sa;
  SECURITY_DESCRIPTOR sd;
  char command_line[256];
  char line[1024];
  char *p;
  int oldversion;
  extern struct options o;

  CLEAR (start_info);
  CLEAR (proc_info);
  CLEAR (sa);
  CLEAR (sd);
   
  exit_event = CreateEvent (NULL, TRUE, FALSE, "openvpn_exit");
  if (exit_event == NULL)
    {
#ifdef DEBUG
      PrintErrorDebug("CreateEvent(openvpn_exit) failed.");
#endif
      ShowMsg(GUI_NAME, "CreateEvent() failed when checking openvpn version.");
      return(false);
    }

#ifdef DEBUG
  PrintErrorDebug("CreateEvent(openvpn_exit) succeded.");
#endif

  /* construct command line */
  mysnprintf (command_line, "openvpn --version");

  /* Make security attributes struct for logfile handle so it can
     be inherited. */
  sa.nLength = sizeof (sa);
  sa.lpSecurityDescriptor = &sd;
  sa.bInheritHandle = TRUE;
  if (!InitializeSecurityDescriptor (&sd, SECURITY_DESCRIPTOR_REVISION))
    {
      ShowMsg (GUI_NAME, "InitializeSecurityDescriptor failed");
      return(0);
    }
  if (!SetSecurityDescriptorDacl (&sd, TRUE, NULL, FALSE))
    {
      ShowMsg (GUI_NAME, "SetSecurityDescriptorDacl failed");
      return(0);
    }

  /* Create the child input pipe. */
  if (!CreatePipe(&hInputRead,&hInputWriteTmp,&sa,0))
    {
      ShowMsg(GUI_NAME, "CreatePipe on hInputRead failed.");
      return(0);
    }

  /* Create the child output pipe. */
  if (!CreatePipe(&hOutputReadTmp,&hOutputWrite,&sa,0))
    {
      ShowMsg(GUI_NAME, "CreatePipe on hOutputWrite failed.");
      return(0);
    }

  if (!DuplicateHandle(GetCurrentProcess(),hOutputReadTmp,
                       GetCurrentProcess(),
                       &hOutputRead, // Address of new handle.
                       0,FALSE, // Make it uninheritable.
                       DUPLICATE_SAME_ACCESS))
    {
      ShowMsg(GUI_NAME, "DupliateHandle on hOutputRead failed.");
      return(0);
    }

  if (!DuplicateHandle(GetCurrentProcess(),hInputWriteTmp,
                       GetCurrentProcess(),
                       &hInputWrite, // Address of new handle.
                       0,FALSE, // Make it uninheritable.
                       DUPLICATE_SAME_ACCESS))
    {
      ShowMsg(GUI_NAME, "DupliateHandle on hInputWrite failed.");
      return(0);
    }


  /* Close inheritable copies of the handles */
  if (!CloseHandle(hOutputReadTmp) || !CloseHandle(hInputWriteTmp)) 
    {
      ShowMsg(GUI_NAME, "CloseHandle on hOutputReadTmp/hInputWriteTmp failed.");
      return(0); 
    }

  /* fill in STARTUPINFO struct */
  GetStartupInfo(&start_info);
  start_info.cb = sizeof(start_info);
  start_info.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
  start_info.wShowWindow = SW_HIDE;
  start_info.hStdInput = hInputRead;
  start_info.hStdOutput = hOutputWrite;
  start_info.hStdError = hOutputWrite;

  /* Start OpenVPN to check version */
  if (!CreateProcess(o.exe_path,
		     command_line,
		     NULL,
		     NULL,
		     TRUE,
		     CREATE_NEW_CONSOLE,
		     NULL,
		     o.config_dir,
		     &start_info,
		     &proc_info))
    {
      ShowMsg (GUI_NAME, "CreateProcess failed, exe='%s' cmdline='%s' dir='%s'",
          o.exe_path,
          command_line,
          o.config_dir);
      return(0);
    }

  /* Default value for oldversion */
  oldversion=0;

  /* Default string to look for to report "Connected". */
  strncpy(o.connect_string, "Successful ARP Flush", sizeof(o.connect_string));

  if (ReadLineFromStdOut(hOutputRead, 0, line) == 1)
    {
#ifdef DEBUG
      PrintDebug("VersionString: %s", line);
#endif
      if (line[8] == '2') /* Majorversion = 2 */
        {
          if (line[10] == '0') /* Minorversion = 0 */
            {
              if ((p=strstr(line, "beta")) != NULL)
                {
                  if (p[5] == ' ') /* 2.0-beta1 - 2.0-beta9 */
                    {
                      if (p[4] >= '6') /* 2.0-beta6 - 2.0-beta9 */
                        {
                          oldversion=0;
                        }
                      else /* < 2.0-beta6 */
                        {
                          oldversion=1;
                        }
                    }
                  else /* >= 2.0-beta10 */
                    {
                      if (strncmp(&p[6], "ms", 2) == 0) /* 2.0-betaXXms */
                        strncpy(o.connect_string, "Initialization Sequence Completed",
                                sizeof(o.connect_string));
                      if ( !((p[4] == 1) && (p[5] == 0)) ) /* >= 2.0-beta11 */
                        strncpy(o.connect_string, "Initialization Sequence Completed",
                                sizeof(o.connect_string));
                        
                      oldversion=0;
                    }
                }
              else /* 2.0 non-beta */
                {
                  strncpy(o.connect_string, "Initialization Sequence Completed",
                          sizeof(o.connect_string));
                  oldversion=0;
                }
            }
          else /* > 2.0 */
            {
              strncpy(o.connect_string, "Initialization Sequence Completed",
                      sizeof(o.connect_string));
              oldversion=0;
            }
        }
      else
        {
          if (line[8] == '1') /* Majorversion = 1 */
            {
              oldversion=1;
            }
          else /* Majorversion != (1 || 2) */
            {
              oldversion=0;
            }
        }
    }
  else return(0);

  o.oldversion = oldversion;
  

  if(!CloseHandle (proc_info.hThread) || !CloseHandle (hOutputWrite)
     || !CloseHandle (hInputRead) || !CloseHandle(exit_event))
    {
      ShowMsg (GUI_NAME, "CloseHandle failed");
      return(0);
    }

  return(1); 
}

/* Return num of connections with Status = CheckVal */
int CountConnectedState(int CheckVal)
{
  int i;
  int count=0;
  extern struct options o;
  
  for (i=0; i < o.num_configs; i++)
    {
      if (o.cnn[i].connect_status == CheckVal)
        count++;
    }

  return (count);
}

void CheckAndSetTrayIcon()
{

  /* Change tray icon if no more connections is running */
  if (CountConnectedState(2) != 0)
    SetTrayIcon(2);
  else
    {
      if ((CountConnectedState(1) != 0) || (CountConnectedState(3) != 0))
        SetTrayIcon(1);
      else
        SetTrayIcon(0);
    }
}

void ThreadOpenVPNStatus(int config)
{
  char conn_name[200];
  HANDLE hThread; 
  DWORD IDThread; 
  extern struct options o;
  char msg[200];
  MSG messages;

  /* Cut of extention from config filename. */
  strncpy(conn_name, o.cnn[config].config_file, sizeof(conn_name));
  conn_name[strlen(conn_name) - (strlen(o.ext_string)+1)]=0;

  /* Show Status Dialog */  
  if (!(o.cnn[config].hwndStatus = CreateDialog (o.hInstance, 
        MAKEINTRESOURCE (IDD_STATUS),
        NULL, (DLGPROC) StatusDialogFunc)))
    ExitThread(1);
  SetDlgItemText(o.cnn[config].hwndStatus, TEXT_STATUS, "Current State: Connecting"); 
  SetDlgItemInt(o.cnn[config].hwndStatus, TEXT_CONFIG, config, FALSE);
  mysnprintf(msg, "OpenVPN Connection (%s)", conn_name);
  SetWindowText(o.cnn[config].hwndStatus, msg);
  ShowWindow(o.cnn[config].hwndStatus, SW_SHOW);

  /* Start Thread to monitor our OpenVPN process */
  hThread = CreateThread(NULL, 0, 
            (LPTHREAD_START_ROUTINE) WatchOpenVPNProcess,
            (int *) config,  // pass config nr
            0, &IDThread); 
  if (hThread == NULL) 
    {
      ShowMsg (GUI_NAME, "CreateThread to read openvpn process stdout Failed.");
      ExitThread(0);
    }

  /* Run the message loop. It will run until GetMessage() returns 0 */
  while (GetMessage (&messages, NULL, 0, 0))
    {
      if(!IsDialogMessage(o.cnn[config].hwndStatus, &messages))
      {
        TranslateMessage(&messages);
        DispatchMessage(&messages);
      }
    }
  ExitThread(0);

}
